#include "ns3/command-line.h"
#include "ns3/config.h"
#include "ns3/double.h"
#include "ns3/internet-stack-helper.h"
#include "ns3/ipv4-address-helper.h"
#include "ns3/ipv4-list-routing-helper.h"
#include "ns3/ipv4-static-routing-helper.h"
#include "ns3/log.h"
#include "ns3/mobility-helper.h"
#include "ns3/mobility-model.h"
#include "ns3/olsr-helper.h"
#include "ns3/string.h"
#include "ns3/uinteger.h"
#include "ns3/yans-wifi-channel.h"
#include "ns3/yans-wifi-helper.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/flow-monitor-helper.h"
#include "ns3/netanim-module.h"
using namespace ns3;

NS_LOG_COMPONENT_DEFINE("WifiSimpleAdhocGrid");

void
ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        
    }
}

static void
GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize, uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet>(pktSize));
        Simulator::Schedule(pktInterval,
                            &GenerateTraffic,
                            socket,
                            pktSize,
                            pktCount - 1,
                            pktInterval);
    }
   else
    {
        socket->Close();
    }
}

int
main(int argc, char* argv[])
{
    std::string phyMode("DsssRate1Mbps");
    double distance = 100;      // m
    uint32_t packetSize = 1000; // bytes
    uint32_t numPackets = 100;
   
   
    uint32_t packetSize2 = 1000; // bytes
    uint32_t numPackets2 = 225;
    
    uint32_t packetSize3 = 1200; // bytes
    uint32_t numPackets3 = 532;
    
    uint32_t packetSize4 = 1500; // bytes
    uint32_t numPackets4 = 165;
    
    uint32_t numNodes = 4; // by default, 5x5
    uint32_t sinkNode = 0;
    uint32_t sourceNode = 3;
    
     uint32_t sinkNode1 = 1;
    uint32_t sourceNode1 = 2;
    
    uint32_t sinkNode2 = 2;
    uint32_t sourceNode2 = 0;
      
    uint32_t sinkNode3 = 3;
    uint32_t sourceNode3 = 1;
    
    double interval = 0.004; // seconds
    bool verbose = false;
    bool tracing = true;

    CommandLine cmd(__FILE__);
    cmd.AddValue("phyMode", "Wifi Phy mode", phyMode);
    cmd.AddValue("distance", "distance (m)", distance);
    cmd.AddValue("packetSize", "size of application packet sent", packetSize);
    cmd.AddValue("numPackets", "number of packets generated", numPackets);
    cmd.AddValue("interval", "interval (seconds) between packets", interval);
    cmd.AddValue("verbose", "turn on all WifiNetDevice log components", verbose);
    cmd.AddValue("tracing", "turn on ascii and pcap tracing", tracing);
    cmd.AddValue("numNodes", "number of nodes", numNodes);
    cmd.AddValue("sinkNode", "Receiver node number", sinkNode);
    cmd.AddValue("sourceNode", "Sender node number", sourceNode);
    cmd.Parse(argc, argv);
    // Convert to time object
    Time interPacketInterval = Seconds(interval);

    // Fix non-unicast data rate to be the same as that of unicast
    Config::SetDefault("ns3::WifiRemoteStationManager::NonUnicastMode", StringValue(phyMode));

    NS_LOG_UNCOND("Topology");
    NS_LOG_UNCOND("-----------");
    NS_LOG_UNCOND("|n0  |  n1|");
    NS_LOG_UNCOND("-----------");
    NS_LOG_UNCOND("|n2  |  n3|");
    NS_LOG_UNCOND("-----------");
    NS_LOG_UNCOND("");
    NS_LOG_UNCOND("-----------------------------------------------------------");
    NS_LOG_UNCOND("UDP Flow Configurations");
    NS_LOG_UNCOND("n3 -----> n0");
    NS_LOG_UNCOND("n2 -----> n1");
    NS_LOG_UNCOND("n0 -----> n2");
    NS_LOG_UNCOND("n1 -----> n3");
    NS_LOG_UNCOND("-----------------------------------------------------------");
    
    
    NodeContainer c;
    c.Create(numNodes);

    // The below set of helpers will help us to put together the wifi NICs we want
    WifiHelper wifi;
    if (verbose)
    {
        WifiHelper::EnableLogComponents(); // Turn on all Wifi logging
    }
   wifi.SetStandard(WIFI_STANDARD_80211b);
    YansWifiPhyHelper wifiPhy;
    // set it to zero; otherwise, gain will be added
    wifiPhy.Set("RxGain", DoubleValue(0));
     
    // ns-3 supports RadioTap and Prism tracing extensions for 802.11b
    wifiPhy.SetPcapDataLinkType(WifiPhyHelper::DLT_IEEE802_11_RADIO);

    YansWifiChannelHelper wifiChannel;
    wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChannel.AddPropagationLoss("ns3::FriisPropagationLossModel");
    wifiPhy.SetChannel(wifiChannel.Create());

    // Add an upper mac and disable rate control
    WifiMacHelper wifiMac;
   
    wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode",
                                 StringValue(phyMode),
                                 "ControlMode",
                                 StringValue(phyMode));
    // Set it to adhoc mode
    wifiMac.SetType("ns3::AdhocWifiMac");
    NetDeviceContainer devices = wifi.Install(wifiPhy, wifiMac, c);

    MobilityHelper mobility;
    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
                                  "MinX",
                                  DoubleValue(0.0),
                                  "MinY",
                                  DoubleValue(0.0),
                                  "DeltaX",
                                  DoubleValue(distance),
                                  "DeltaY",
                                  DoubleValue(distance),
                                  "GridWidth",
                                  UintegerValue(2),
                                  "LayoutType",
                                  StringValue("RowFirst"));
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(c);

    // Enable OLSR
 OlsrHelper olsr; 
   
    Ipv4ListRoutingHelper list;
  list.Add(olsr, 0);

    InternetStackHelper internet;
    internet.SetRoutingHelper(list); 
    internet.Install(c);

    NS_LOG_INFO("Assign IP Addresses.");
    Ipv4AddressHelper ipv4;
    
    ipv4.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer i = ipv4.Assign(devices);
    

//udp flow-1:

    TypeId tid = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink = Socket::CreateSocket(c.Get(sinkNode), tid);
    InetSocketAddress local = InetSocketAddress(Ipv4Address::GetAny(), 80);
    recvSink->Bind(local);
    recvSink->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source = Socket::CreateSocket(c.Get(sourceNode), tid);
    InetSocketAddress remote = InetSocketAddress(i.GetAddress(sinkNode, 0), 80);
    source->Connect(remote);

//udp flow-2:
    TypeId tid1 = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink1 = Socket::CreateSocket(c.Get(sinkNode1), tid1);
    InetSocketAddress local1 = InetSocketAddress(Ipv4Address::GetAny(), 82);
    recvSink1->Bind(local1);
    recvSink1->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source1 = Socket::CreateSocket(c.Get(sourceNode1), tid1);
    InetSocketAddress remote1 = InetSocketAddress(i.GetAddress(sinkNode1, 0), 82);
    source1->Connect(remote1);

//udp flow-3:
    TypeId tid2 = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink2 = Socket::CreateSocket(c.Get(sinkNode2), tid2);
    InetSocketAddress local2 = InetSocketAddress(Ipv4Address::GetAny(), 91);
    recvSink2->Bind(local2);
    recvSink2->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source2 = Socket::CreateSocket(c.Get(sourceNode2), tid2);
    InetSocketAddress remote2 = InetSocketAddress(i.GetAddress(sinkNode2, 0), 82);
    source2->Connect(remote2);

//udp flow-4:
    TypeId tid3 = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink3 = Socket::CreateSocket(c.Get(sinkNode3), tid3);
    InetSocketAddress local3 = InetSocketAddress(Ipv4Address::GetAny(), 94);
    recvSink3->Bind(local3);
    recvSink3->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source3 = Socket::CreateSocket(c.Get(sourceNode3), tid3);
    InetSocketAddress remote3 = InetSocketAddress(i.GetAddress(sinkNode3, 0), 82);
    source3->Connect(remote3);   
    if (tracing == true)
    {
        AsciiTraceHelper ascii;
        wifiPhy.EnableAsciiAll(ascii.CreateFileStream("exp-10-final.tr"));
        wifiPhy.EnablePcap("wifi-simple-adhoc-grid", devices);
        
    }

    
    Simulator::Schedule(Seconds(10.0),
                        &GenerateTraffic,
                        source,
                        packetSize,
                        numPackets,
                        interPacketInterval);
 Simulator::Schedule(Seconds(20.0),
                        &GenerateTraffic,
                        source1,
                        packetSize2,
                        numPackets2,
                        interPacketInterval);
                        
Simulator::Schedule(Seconds(30.0),
                        &GenerateTraffic,
                        source2,
                        packetSize3,
                        numPackets3,
                        interPacketInterval);
   Simulator::Schedule(Seconds(40.0),
                        &GenerateTraffic,
                        source3,
                        packetSize4,
                        numPackets4,
                        interPacketInterval);
 
    Simulator::Stop(Seconds(100.0));
    //flow monitor
    Ptr<FlowMonitor> flowMonitor;
    FlowMonitorHelper flowHelper;
    flowMonitor= flowHelper.InstallAll();
    
    
     AnimationInterface anim ("animationwifi-adhoc-wireless.xml");
     
     NS_LOG_INFO ("Run Simulation.");
     Simulator::Run();
    
    Ptr<Ipv4FlowClassifier> classifier=DynamicCast<Ipv4FlowClassifier> (flowHelper.GetClassifier());
    std::map<FlowId,FlowMonitor::FlowStats> stats=flowMonitor->GetFlowStats();
     

    for(std::map<FlowId,FlowMonitor::FlowStats>::const_iterator iter=stats.begin();iter!=stats.end();++iter)
    {
    Ipv4FlowClassifier::FiveTuple t =classifier->FindFlow(iter->first);
    NS_LOG_UNCOND("Flow ID:- "<< iter->first<< " Source addr: "<< t.sourceAddress << " Dest Addr: "<<t.destinationAddress ); 
     uint32_t protocol = t.protocol;
    if (protocol == 6) {
        std::cout << "Type: TCP Flow" << std::endl;
    } else if (protocol == 17) {
        std::cout << "Type: UDP Flow" << std::endl;
    } 
    	  NS_LOG_UNCOND("Tx Packets = " << iter->second.txPackets);
    	  NS_LOG_UNCOND("Rx Packets = " << iter->second.rxPackets);
    	  NS_LOG_UNCOND("Lost packets = " << iter->second.lostPackets);
    	  NS_LOG_UNCOND("Times forwarded = " << iter->second.timesForwarded);
    	  NS_LOG_UNCOND("Delay = " << iter->second.delaySum);
    	  NS_LOG_UNCOND("Throughput: " << iter->second.rxBytes * 8.0 / (iter->second.timeLastRxPacket.GetSeconds()-iter->second.timeFirstTxPacket.GetSeconds()) / 1024  << " Kbps");
   	NS_LOG_UNCOND("-----------------------------------------------------------");
    }
   
    Simulator::Destroy();

    return 0;
}
